#include "license_env.h"
#include "xalloc.h"
#include "license_logging.h"

#include <stdlib.h>
#include <string.h>

struct op5lic_env_list_t {
	unsigned int len;
	struct op5lic_env_node_t **nodes;
};

void op5lic_env_free_tokenized_path(char **path, int depth) {
	int i;
	for (i = 0; i < depth; ++i) {
		xfree(path[i]);
	}
	xfree(path);
}

char ** op5lic_env_tokenize_path(const char *nodepath, int * depth) {
	if (!nodepath)
		return NULL;
	char *k = xstrdup(nodepath);
	char ** path = NULL;
	char * token = NULL;
	char * tmp = k;

	*depth = 0;
	log_debug("Tokenizing nodepath: %s", nodepath);
	while (NULL != (token = strsep(&k, "."))) {
		if (0 == strcmp(token, ""))
			continue;

		log_debug("Found token: %s", token);
		++(*depth);
		path = xrealloc(path, ((*depth) * sizeof(char *)));
		path[(*depth)-1] = xstrdup(token);
	}
	xfree(tmp);
	return path;
}

struct op5lic_env_list_t *op5lic_env_node_to_list(struct op5lic_env_node_t *node) {
	if (!node)
		return NULL;

	struct op5lic_env_list_t *list = NULL;
	if (node->type == ENV_TYPE_ENV_LIST) {
		list = node->value.val_env_list;
		node->type = ENV_TYPE_INTEGER; /* integer doesn't need any cleanup*/
	}

	op5lic_env_node_destroy(node);
	return list;
}

char * op5lic_env_list_to_string(const struct op5lic_env_list_t *list) {
	unsigned int i;
	char * envstr = NULL, *tmp = NULL;
	char *subenvstr = NULL;
	if (!list) {
		envstr = xstrdup("<undefined>");
	}
	else {
		for (i = 0; i < list->len; i++) {
			tmp = envstr;
			xasprintf(&envstr, "%s%s", envstr ? envstr : "", list->nodes[i]->key);
			xfree(tmp);
			tmp = envstr;
			switch (list->nodes[i]->type) {
				case ENV_TYPE_ENV_LIST:
					subenvstr = op5lic_env_list_to_string(list->nodes[i]->value.val_env_list);
					xasprintf(&envstr, "%s -> [ %s ], ", envstr, subenvstr);
					xfree(subenvstr);
					break;
				case ENV_TYPE_STRING:
					xasprintf(&envstr, "%s -> %s, ", envstr, list->nodes[i]->value.val_string);
					break;
				case ENV_TYPE_INTEGER:
					xasprintf(&envstr, "%s -> %d, ", envstr, list->nodes[i]->value.val_int);
					break;
			}
			xfree(tmp);
		}
	}
	return envstr;
}

struct op5lic_env_node_t *op5lic_env_list_find(const struct op5lic_env_list_t *list, const char *nodepath) {
	int depth, i;
	char **path = NULL;
	char *path_elem;

	if (!list || !nodepath)
		return NULL;

	struct op5lic_env_node_t * found_node = NULL, * node = NULL;

	path = op5lic_env_tokenize_path(nodepath, &depth);
	for (i = 0; i < depth && found_node == NULL; ++i) {
		path_elem = path[i];
		unsigned int j;
		for (j = 0; j < list->len; ++j) {
			node = list->nodes[j];
			if (0 == strcmp(path_elem, node->key)) {
				if (i == depth - 1) {
					/* we've reached the desired depth, and the name matches:
					 * this is the node we're looking for (regardless of node type).
					 */
					found_node = node;
				} else if (node->type == ENV_TYPE_ENV_LIST){
					/* we've still got some way to go, but this node is part of the
					 * path to get there */
					list = node->value.val_env_list;
				}
				else {
					/* correct name, but we can't go any further down. Bail! */
					return NULL;
				}
				break;
			}
		}
	}
	op5lic_env_free_tokenized_path(path, depth);
	return found_node;
}

struct op5lic_env_list_t *op5lic_env_list_create(void) {
	struct op5lic_env_list_t *list = xcalloc(1, sizeof(struct op5lic_env_list_t));
	list->nodes= NULL;
	list->len = 0;
	return list;
}

void op5lic_env_list_destroy(struct op5lic_env_list_t *list) {
	unsigned int i;
	if (!list)
		return;

	for (i = 0; i < list->len; ++i) {
		op5lic_env_node_destroy(list->nodes[i]);
	}
	xfree(list->nodes);
	xfree(list);
}


static struct op5lic_env_node_t * op5lic_env_create(const char *key) {
	if (!key)
		return NULL;
	struct op5lic_env_node_t *node;
	if ( strchr(key, '.') != NULL) {
		log_err("Refusing to create environment node with key \"%s\" containing '.' to avoid ambiguity", key);
		return NULL;
	}
	node = xcalloc(1, sizeof(struct op5lic_env_node_t));
	node->key = xstrdup(key);
	return node;
}

struct op5lic_env_node_t * op5lic_env_create_int(const char *key, int v) {
	struct op5lic_env_node_t *node = op5lic_env_create(key);
	if (node == NULL)
		return NULL;

	node->type = ENV_TYPE_INTEGER;
	node->value.val_int = v;
	return node;
}

struct op5lic_env_node_t * op5lic_env_create_list(const char *key) {
	struct op5lic_env_node_t *node = op5lic_env_create(key);
	if (node == NULL)
		return NULL;
	node->type = ENV_TYPE_ENV_LIST;
	node->value.val_env_list = op5lic_env_list_create();
	return node;
}

struct op5lic_env_node_t * op5lic_env_create_string(const char *key, const char *v) {
	struct op5lic_env_node_t *node = op5lic_env_create(key);
	if (node == NULL)
		return NULL;
	node->type = ENV_TYPE_STRING;
	node->value.val_string = xstrdup(v);
	return node;
}

void op5lic_env_node_destroy(struct op5lic_env_node_t *node) {
	if (!node)
		return;

	switch (node->type) {
		case ENV_TYPE_ENV_LIST:
			op5lic_env_list_destroy(node->value.val_env_list);
			break;
		case ENV_TYPE_INTEGER: break;
		case ENV_TYPE_STRING:
			xfree(node->value.val_string);
			break;
	}

	xfree(node->key);
	xfree(node);
}


void op5lic_env_list_add(struct op5lic_env_list_t *list, struct op5lic_env_node_t *new_node) {
	if (new_node == NULL) {
		log_err("Refusing to add Null node to environment list");
		return;
	}
	++(list->len);
	list->nodes = xrealloc(list->nodes, list->len * sizeof(struct op5lic_env_node_t));
	list->nodes[list->len - 1] =  new_node;
}
