#include <string.h>
#include <assert.h>
#include "xalloc.h"
#include "license_metadata.h"
#include "license_logging.h"

struct op5lic_metadata_field_t {
	enum MetadataFieldType type;
	union
	{
		char *string;
		struct op5lic_metadata_t *table;
	} val;
	char *name;

};

struct op5lic_metadata_t {
	size_t nfields;
	struct op5lic_metadata_field_t **fields;
};

struct op5lic_metadata_real_iter_t {
	size_t pos;
	struct op5lic_metadata_t *iter_target;
	struct op5lic_metadata_field_t *curfield;
};

static void op5lic_metadata_field_destroy(struct op5lic_metadata_field_t *);
static struct op5lic_metadata_field_t * op5lic_metadata_get_field(struct op5lic_metadata_t *md, enum MetadataFieldType type, const char *fieldname) {
	size_t i;
	for (i = 0; i < md->nfields; ++i) {
		if (md->fields[i]->type == type) {
			if (0 == strcmp(fieldname, md->fields[i]->name)) {
				return md->fields[i];
			}
		}
	}
	return NULL;
}

struct op5lic_metadata_t *op5lic_metadata_get_table(struct op5lic_metadata_t *md, const char *fieldname) {
	struct op5lic_metadata_field_t *mdf;
	mdf = op5lic_metadata_get_field(md, MD_TYPE_TABLE, fieldname);
	if (!mdf)
		return NULL;

	return mdf->val.table;
}

const char * op5lic_metadata_get_str(struct op5lic_metadata_t *md, const char *fieldname) {
	struct op5lic_metadata_field_t *mdf;
	mdf = op5lic_metadata_get_field(md, MD_TYPE_STRING, fieldname);
	if (!mdf)
		return NULL;

	return mdf->val.string;
}

void op5lic_metadata_destroy(struct op5lic_metadata_t *md) {
	size_t i;
	for (i = 0; i < md->nfields; ++i) {
		op5lic_metadata_field_destroy(md->fields[i]);
	}
	xfree(md->fields);
	xfree(md);
}

struct op5lic_metadata_field_t * op5lic_metadata_field_create(void) {
	struct op5lic_metadata_field_t *mdf;
	mdf = xcalloc(1, sizeof(*mdf));
	return mdf;
}

static void op5lic_metadata_field_destroy(struct op5lic_metadata_field_t *mdf) {
	if (!mdf)
		return;

	switch (mdf->type) {
	case MD_TYPE_TABLE:
		op5lic_metadata_destroy(mdf->val.table);
		break;
	case MD_TYPE_STRING:
		xfree(mdf->val.string);
		break;
	}
	xfree(mdf->name);
	xfree(mdf);
}

struct op5lic_metadata_t * op5lic_metadata_create(void) {
	struct op5lic_metadata_t *md;
	md = xcalloc(1, sizeof(*md));
	md->fields = NULL;
	md->nfields = 0;
	return md;
}

void op5lic_metadata_add_field_str(struct op5lic_metadata_t * md,
		const char *fieldname, const char *value) {
	struct op5lic_metadata_field_t * field = op5lic_metadata_field_create();

	field->type = MD_TYPE_STRING;
	field->name = xstrdup(fieldname);
	field->val.string = xstrdup(value);
	++md->nfields;
	md->fields = xrealloc(md->fields, (md->nfields * sizeof(*field)));
	md->fields[md->nfields-1] = field;
}

void op5lic_metadata_add_field_table(struct op5lic_metadata_t * md,
		const char *fieldname, struct op5lic_metadata_t *tbl) {
	struct op5lic_metadata_field_t * field = op5lic_metadata_field_create();
	field->type = MD_TYPE_TABLE;
	field->name = xstrdup(fieldname);
	field->val.table = tbl;
	++md->nfields;
	md->fields = xrealloc(md->fields, (md->nfields * sizeof(*field)));
	md->fields[md->nfields-1] = field;
}

struct op5lic_metadata_iter_t * op5lic_metadata_iter_next(struct op5lic_metadata_iter_t *iter, char **key, void **value, enum MetadataFieldType *type) {
	struct op5lic_metadata_real_iter_t *it = (struct op5lic_metadata_real_iter_t *)iter;
	if (!it|| !it->iter_target)
		return NULL;

	++(it->pos);
	if (it->pos >= it->iter_target->nfields) {
		it->curfield = NULL;
		it->pos = -1;
		*key = NULL;
		*value = NULL;
		return NULL;
	}

	it->curfield = it->iter_target->fields[it->pos];
	*key = it->curfield->name;
	*type = it->curfield->type;
	switch (*type) {
	case MD_TYPE_TABLE:
			*value = it->curfield->val.table;
	case MD_TYPE_STRING:
			*value = it->curfield->val.string;
	}
	return (struct op5lic_metadata_iter_t *)it;
}

const char * op5lic_metadata_iter_key(struct op5lic_metadata_iter_t *iter) {
	struct op5lic_metadata_real_iter_t *it = (struct op5lic_metadata_real_iter_t *) iter;
	return (const char *)it->curfield->name;
}

void * op5lic_metadata_iter_value(struct op5lic_metadata_iter_t *iter) {
	struct op5lic_metadata_real_iter_t *it = (struct op5lic_metadata_real_iter_t *) iter;
	switch (it->curfield->type) {
		case MD_TYPE_TABLE:
			return it->curfield->val.table;
		case MD_TYPE_STRING:
			return it->curfield->val.string;
	}
	log_err("BUG!!! Impossible value for metadata field type!");
	assert(0);
}

enum MetadataFieldType op5lic_metadata_iter_type(struct op5lic_metadata_iter_t *iter) {
	struct op5lic_metadata_real_iter_t *it = (struct op5lic_metadata_real_iter_t *) iter;
	return it->curfield->type;
}

void op5lic_metadata_iter_init(struct op5lic_metadata_iter_t *iter, struct op5lic_metadata_t * md) {
	struct op5lic_metadata_real_iter_t *it = (struct op5lic_metadata_real_iter_t *) iter;
	it->iter_target = md;
	it->pos = -1;
	it->curfield = NULL;
}

